"""
Time projectors

The classes in this module are used to infer values for a scenario at later times given
the trends before that time.

"""

from .extend_latest_time_quantile import ExtendLatestTimeQuantile  # noqa: F401
from .extend_rms_closest import ExtendRMSClosest  # noqa: F401
from .linear_extender import LinearExtender  # noqa: F401



"""
Module for the database cruncher which uses the 'closest RMS' technique.
"""
import warnings

import numpy as np
import pandas as pd
import pyam

from .base import _DatabaseCruncher


class RMSClosest(_DatabaseCruncher):
    """
    Database cruncher which uses the 'closest RMS' technkque.

    This cruncher derives the relationship between two or more variables by finding the
    scenario which has the most similar timeseries for the lead gases in the database.
    The follower gas timeseries is then simply copied from the closest scenario.

    Here, 'most similar' is defined as the smallest time-averaged root mean squared (L2)
    difference. If multiple lead values are used, they may be weighted differently to
    account for differences between the reported units. The most similar model/scenario
    combination minimises

    .. math::
        RMS error = \\sum_l w_l \\left ( \\frac{1}{n} \\sum_{t=0}^n (E_l(t) - e_l(t))^2 \\right )^{1/2}

    where :math:`l` is a lead gas, :math:`w_l` is a weighting for that lead gas,
    :math:`n` is the total number of timesteps in all lead gas timeseries,
    :math:`E_l(t)` is the lead gas emissions timeseries and :math:`e_l(t)` is a lead
    gas emissions timeseries in the infiller database.
    """

    def derive_relationship(self, variable_follower, variable_leaders, weighting=None):
        """
        Derive the relationship between the lead and the follow variables from the
        database.

        Parameters
        ----------
        variable_follower : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|C5F12"``).

        variable_leaders : list[str]
            The variable we want to use in order to infer timeseries of
            ``variable_follower`` (e.g. ``["Emissions|CO2"]``). This may contain
            multiple elements.

        weighting : dict{str: float}
            When used with multiple lead variables, this weighting factor controls the
            relative importance of different variables for determining closeness. E.g.
            if wanting to compare both CO2 and CH4 emissions reported in mass
            units but weighted by the AR5 GWP100 metric, this would be
            {"Emissions|CO2": 1, "Emissions|CH4": 28}.

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable_leaders`` timeseries and returns timeseries for
            ``variable_follower`` based on the derived relationship between the two.
            Please see the source code for the exact definition (and docstring) of the
            returned function.

        Raises
        ------
        ValueError
            ``variable_leaders`` contains more than one variable.

        ValueError
            There is no data for ``variable_leaders`` or ``variable_follower`` in the
            database.
        """
        self._check_iamdf_lead(variable_leaders)
        iamdf_follower = self._get_iamdf_section(variable_follower)
        data_follower_time_col = iamdf_follower.time_col
        iamdf_lead = self._db.filter(variable=variable_leaders)
        if not weighting:
            weighting = {variab: 1 for variab in variable_leaders}
        if any(var not in weighting.keys() for var in variable_leaders):
            raise ValueError("Weighting does not include all lead variables.")
        iamdf_lead, iamdf_follower = _filter_for_overlap(
            iamdf_lead,
            iamdf_follower,
            ["scenario", "model", data_follower_time_col],
            variable_leaders,
        )

        leader_var_unit = {
            var["variable"]: var["unit"]
            for _, var in iamdf_lead[["variable", "unit"]].drop_duplicates().iterrows()
        }

        def filler(in_iamdf):
            """
            Filler function derived from :obj:`RMSClosest`.

            Parameters
            ----------
            in_iamdf : :obj:`pyam.IamDataFrame`
                Input data to fill data in

            Returns
            -------
            :obj:`pyam.IamDataFrame`
                Filled in data (without original source data)

            Raises
            ------
            ValueError
                If there are any inconsistencies between the timeseries, units or
                expectations of the program and ``in_iamdf``, compared to the database
                used to generate this ``filler`` function.
            """
            lead_var = in_iamdf.filter(variable=variable_leaders)

            var_units = lead_var[["variable", "unit"]].drop_duplicates()
            if any([key not in lead_var.variable for key in leader_var_unit.keys()]):
                raise ValueError(
                    "Not all required variables are present in the infillee database"
                )
            if any(
                unit["unit"] != leader_var_unit[unit["variable"]]
                for _, unit in var_units.iterrows()
            ):
                raise ValueError(
                    "Units of lead variable is meant to be {}, found {}".format(
                        leader_var_unit, var_units
                    )
                )

            if data_follower_time_col != in_iamdf.time_col:
                raise ValueError(
                    "`in_iamdf` time column must be the same as the time column used "
                    "to generate this filler function (`{}`)".format(
                        data_follower_time_col
                    )
                )

            key_timepoints_filter_iamdf = {
                data_follower_time_col: list(set(lead_var[data_follower_time_col]))
            }
            key_timepoints_filter_lead = {
                data_follower_time_col: list(set(iamdf_lead[data_follower_time_col]))
            }

            def get_values_at_key_timepoints(idf, time_filter):
                # filter warning about empty data frame as we handle it ourselves
                to_return = idf.filter(**time_filter)
                if to_return.data.empty:
                    raise ValueError(
                        "No time series overlap between the original and unfilled data"
                    )
                return to_return

            lead_var_filt = get_values_at_key_timepoints(
                lead_var, key_timepoints_filter_lead
            )
            lead_var_timeseries = lead_var_filt.timeseries()
            iamdf_lead_timeseries = get_values_at_key_timepoints(
                iamdf_lead, key_timepoints_filter_iamdf
            ).timeseries()

            output_ts_list = []
            for _, (model, scenario) in (
                lead_var_filt.data[["model", "scenario"]].drop_duplicates().iterrows()
            ):
                lead_var_mod_scen = lead_var_timeseries[
                    (lead_var_timeseries.index.get_level_values("model") == model)
                    & (
                        lead_var_timeseries.index.get_level_values("scenario")
                        == scenario
                    )
                ]
                if len(lead_var_mod_scen) != len(variable_leaders):
                    raise ValueError(
                        "Insufficient variables are found to infill model {}, scenario "
                        "{}. Only found {}.".format(model, scenario, lead_var_mod_scen)
                    )
                closest_model, closest_scenario = _select_closest(
                    iamdf_lead_timeseries, lead_var_mod_scen, weighting
                )

                # Filter to find the matching follow data for the same model, scenario
                # and region
                tmp = iamdf_follower.filter(
                    model=closest_model, scenario=closest_scenario
                ).data

                # Update the model and scenario to match the elements of the input.
                tmp["model"] = model
                tmp["scenario"] = scenario
                output_ts_list.append(tmp)
                for col in in_iamdf.extra_cols:
                    tmp[col] = lead_var_mod_scen.index.get_level_values(col).tolist()[0]
            return pyam.concat(output_ts_list)

        return filler

    def _check_iamdf_lead(self, variable_leaders):
        if not all([v in self._db.variable for v in variable_leaders]):
            error_msg = "No data for `variable_leaders` ({}) in database".format(
                variable_leaders
            )
            raise ValueError(error_msg)

    def _get_iamdf_section(self, variables):
        # filter warning about empty data frame as we handle it ourselves
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            iamdf_section = self._db.filter(variable=variables)

        data_section = iamdf_section.data
        if data_section.empty:
            error_msg = "No data for `variable_follower` ({}) in database".format(
                variables
            )
            raise ValueError(error_msg)

        return iamdf_section


def _select_closest(to_search_df, target_df, weighting):
    """
    Find model/scenario combo in ``to_search_df`` that is closest to that of the target.

    Here, 'closest' is in the root-mean squared sense. In the event that multiple model/
    scenarios are equally close, returns first row.

    Parameters
    ----------
    to_search_df : :obj:`pd.DataFrame`
        The model/scenario combos to search for the closest case. A timeseries.

    target_df : :obj:`pd.DataFrame`
        The data to which we want to be close. A timeseries.

    weighting : map{str: float}
        Maps the variable name onto the weighting for that variable.

    Returns
    -------
    dict
        Index of the closest timeseries.
    """

    if target_df.shape[1] != to_search_df.shape[1]:
        raise ValueError(
            "Target array does not match the size of the searchable arrays"
        )
    if any(
        x not in target_df.index.get_level_values("variable")
        for x in to_search_df.index.get_level_values("variable")
    ):
        raise ValueError("No variable overlap between target and infiller databases.")
    if any(col not in target_df.columns for col in to_search_df.columns):
        raise ValueError("Time values mismatch between target and infiller databases.")
    rms = pd.Series(index=to_search_df.index, dtype=np.float64)
    target_for_var = {}
    for var in to_search_df.index.get_level_values("variable").unique():
        target_for_var[var] = target_df[
            target_df.index.get_level_values("variable") == var
        ].squeeze()
    var_index = to_search_df.index.names.index("variable")
    for label, row in to_search_df.iterrows():
        varname = label[var_index]
        rms.loc[label] = (
            ((target_for_var[varname] - row) ** 2).mean()
        ) ** 0.5 * weighting[varname]
    rmssums = rms.groupby(level=["model", "scenario"], sort=False).sum()
    return rmssums.idxmin()


def _filter_for_overlap(df1, df2, cols, leaders):
    """
    Returns overlapping model/scenario combinations in the two input dataframes, which
    must have the same columns.
    Parameters
    ----------
    df1 : :obj:`pd.DataFrame`
        The first dataframe (order is irrelevant)
    df2 : :obj:`pd.DataFrame`
        The second dataframe (order is irrelevant)
    cols: list[str]
        List of columns that should be identical between the two dataframes. Typically
        "scenario", "model", and whatever the time column is.
    leaders : list[str]
        List of lead variables that must be found in all acceptable model/scenarios
        combinations.
    Returns
    -------
    (:obj:`pd.DataFrame`, :obj:`pd.DataFrame`)
        The two dataframes in the order they were put in, now filtered for some columns
        being identical.
    """
    lead_data = df1.data.set_index(cols)
    follow_data = df2.data.set_index(cols)
    # We only want to select model/scenario cases where we have data for all leaders
    shared_indices = [
        ind
        for ind in follow_data.index
        if lead_data.index.tolist().count(ind) == len(leaders)
    ]
    # We need to remove duplicates
    shared_indices = list(dict.fromkeys(shared_indices))
    if shared_indices:
        lead_data = lead_data.loc[shared_indices]
        follow_data = follow_data.loc[shared_indices]
        return pyam.IamDataFrame(lead_data), pyam.IamDataFrame(follow_data)
    raise ValueError("No model/scenario overlap between leader and follower data")



"""
Module for the database cruncher that uses the rms closest extension method
"""
import logging

import numpy as np
import pandas as pd
from pyam import IamDataFrame

logger = logging.getLogger(__name__)


class ExtendRMSClosest:
    """
    Time projector which extends the timeseries of a variable with
    future timesteps infilled using the values from the 'closest'
    pathway in the infilling database.

    We define the closest pathway as the pathway with the smallest
    time-averaged (over the reported time steps) root mean squared
    difference
    """

    def __init__(self, db):
        """
        Initialise the time projector with a database that contains
        the range of times you wish to see in the output.

        Parameters
        ----------
        db : IamDataFrame
            The database to use
        """
        self._db = db.copy()

    def derive_relationship(self, variable):
        """
        Derives the values for the model/scenario combination in the database
        with the least RMS error.

        Parameters
        ----------
        variable : str
            The variable for which we want to calculate the timeseries (e.g.
            `Emissions|CO2`).

        Returns
        -------
        :obj: `pyam.IamDataFrame`
            Filled in data (without original source data)
        """
        iamdf = self._get_iamdf_variable(variable)

        infiller_time_col = iamdf.time_col
        data_follower_unit = iamdf.data["unit"].unique()

        assert (
            len(data_follower_unit) == 1
        ), "The infiller database has {} units in it. It should have one".format(
            len(data_follower_unit)
        )

        def filler(in_iamdf):
            """
            Filler function

            Parameters
            ----------
            in_iamdf : pyam.IamDataFrame
                Input data to be infilled

            Returns
            -------
            :obj:pyam.IamDataFrame
                Filled in data (without original source data)

            Raises
            ------
            ValueError
                "The infiller database does not extend in time past the target "
                "database, so no infilling can occur."
            """
            target_df = in_iamdf.filter(variable=variable)
            if target_df.empty:
                error_msg = "No data for `variable`({}) in target dataframe".format(
                    variable
                )
                raise ValueError(error_msg)

            key_timepoints = target_df.data[infiller_time_col]
            later_times = [
                t
                for t in iamdf.data[infiller_time_col].unique()
                if t > max(key_timepoints)
            ]

            if not later_times:
                raise ValueError(
                    "The infiller database does not extend in time past the target"
                    "database, so no infilling can occur"
                )
            key_timepoint_filter = {infiller_time_col: key_timepoints}

            def get_values_at_key_timepoints(idf, time_filter):
                to_return = idf.filter(**time_filter)
                if to_return.data.empty:
                    raise ValueError(
                        "Not timeseries overlap between original and unfilled data"
                    )

                return to_return.timeseries()

            infiller_at_key_times = get_values_at_key_timepoints(
                iamdf, key_timepoint_filter
            )
            target_at_key_times = get_values_at_key_timepoints(
                target_df, key_timepoint_filter
            )

            closest_model, closest_scenario = _select_closest(
                infiller_at_key_times, target_at_key_times
            )
            tmp = iamdf.filter(
                model=closest_model, scenario=closest_scenario
            ).timeseries()
            output_ts = target_df.timeseries()
            for time in later_times:
                output_ts[time] = tmp[time].values[0]
            for col in output_ts.columns:
                if col not in later_times:
                    del output_ts[col]
            return IamDataFrame(output_ts)

        return filler

    def _get_iamdf_variable(self, variable):
        if variable not in self._db.variable:
            error_msg = "No data for `variable`({}) in database".format(variable)
            raise ValueError(error_msg)

        return self._db.filter(variable=variable)


def _select_closest(to_search_df, target_df):
    if target_df.shape[1] != to_search_df.shape[1]:
        raise ValueError(
            "Target array does not match the size of the searchable arrays"
        )

    rms = pd.Series(index=to_search_df.index, dtype=np.float64)
    target_for_var = {}
    for var in to_search_df.index.get_level_values("variable").unique():
        target_for_var[var] = target_df[
            target_df.index.get_level_values("variable") == var
        ].squeeze()
    var_index = to_search_df.index.names.index("variable")
    for label, row in to_search_df.iterrows():
        varname = label[var_index]
        rms.loc[label] = (((target_for_var[varname] - row) ** 2).mean()) ** 0.5
    rmssums = rms.groupby(level=["model", "scenario"], sort=False).sum()
    return rmssums.idxmin()



"""
Uses the 'time-dependent ratio' database cruncher designed for constructing an
aggregate variable and breaking this mix into its constituents.
"""

import pyam

from silicone.database_crunchers import TimeDepRatio
from silicone.utils import convert_units_to_MtCO2_equiv


class DecomposeCollectionTimeDepRatio:
    """
    Constructs an aggregate variable and uses the 'time-dependent ratio' technique to
    calculate what this predicts for our database.
    """

    def __init__(self, db):
        """
        Initialises the database to use for infilling.

        Parameters
        ----------
        db : IamDataFrame
            The database for infilling.
        """
        self._db = db.copy()

    def _construct_consistent_values(self, aggregate_name, components, db_to_generate):
        """
        Calculates the sum of the components and creates an IamDataFrame with this
        value under variable type `aggregate_name`.

        Parameters
        ----------
        aggregate_name : str
            The name of the aggregate variable.

        components : [str]
            List of the names of the variables to be summed.

        db_to_generate : :obj:`pyam.IamDataFrame`
            Input data from which to construct consistent values.

        Return
        ------
        :obj:`pyam.IamDataFrame`
            Consistently calculated aggregate data.
        """
        assert (
            aggregate_name not in db_to_generate.variable
        ), "We already have a variable of this name"
        relevant_db = db_to_generate.filter(variable=components)
        units = relevant_db.data["unit"].drop_duplicates().sort_values()
        unit_equivs = units.map(lambda x: x.replace("-equiv", "")).drop_duplicates()
        if len(unit_equivs) == 0:
            raise ValueError(
                "Attempting to construct a consistent {} but none of the components "
                "present".format(aggregate_name)
            )
        elif len(unit_equivs) > 1:
            raise ValueError(
                "Too many units found to make a consistent {}".format(aggregate_name)
            )
        use = (
            relevant_db.data.groupby(
                ["model", "scenario", "region", relevant_db.time_col]
            )
            .agg("sum")
            .reset_index()
        )
        # Units are sorted in alphabetical order so we choose the first to get -equiv
        use["unit"] = units.iloc[0]
        use["variable"] = aggregate_name
        for col in relevant_db.extra_cols:
            use[col] = ""
        return pyam.IamDataFrame(use)

    def _set_of_units_without_equiv(self, df):
        """
        Parameters
        ----------
        df : obj:`pyam.IamDataFrame`
            The dataframe whose units we want

        Returns
        -------
        Set(str)
            The set of units from the dataframe with "-equiv" removed
        """
        return set(df.data["unit"].map(lambda x: x.replace("-equiv", "")))

    def infill_components(
        self,
        aggregate,
        components,
        to_infill_df,
        metric_name="AR5GWP100",
        only_consistent_cases=True,
    ):
        """
        Derive the relationship between the composite variables and their sum, then use
        this to deconstruct the sum.

        Parameters
        ----------
        aggregate : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|CO2"``). Unlike in most crunchers, we do not expect the
            database to already contain this data.

        components : list[str]
            The variables whose sum should be equal to the timeseries of the aggregate
             (e.g. ``["Emissions|CO2|AFOLU", "Emissions|CO2|Energy"]``).

        to_infill_df : :obj:`pyam.IamDataFrame`
            The dataframe that already contains the ``aggregate`` variable, but needs
            the ``components`` to be infilled.

        metric_name : str
            The name of the conversion metric to use. This will usually be
            AR<4/5/6>GWP100.

        only_consistent_cases : bool
            Do we want to only use model/scenario combinations where all aggregate and
            components have data at all times? This will reduce the risk of
            inconsistencies or unevenness in the results, but may reduce the amount of
            data.

        Returns
        -------
        :obj:`pyam.IamDataFrame`
            The infilled data resulting from the calculation.

        Raises
        ------
        ValueError
            There is no data for ``variable_leaders`` or ``variable_follower`` in the
            database.
        """
        assert (
            aggregate in to_infill_df.variable
        ), "The database to infill does not have the aggregate variable"
        assert all(
            y not in components for y in to_infill_df.variable
        ), "The database to infill already has some component variables"
        assert len(to_infill_df.data.columns) == len(self._db.data.columns) and all(
            to_infill_df.data.columns == self._db.data.columns
        ), (
            "The database and to_infill_db fed into this have inconsistent columns, "
            "which will prevent adding the data together properly."
        )
        self._filtered_db = self._db.filter(variable=components)
        if self._filtered_db.empty:
            raise ValueError(
                "Attempting to construct a consistent {} but none of the components "
                "present".format(aggregate)
            )
        if only_consistent_cases:
            # Remove cases with nans at some time.
            consistent_cases = (
                self._filtered_db.filter(
                    **{
                        to_infill_df.time_col: to_infill_df[
                            to_infill_df.time_col
                        ].unique()
                    }
                )
                .timeseries()
                .dropna()
            )
            self._filtered_db = pyam.IamDataFrame(consistent_cases)

        # We only want to reference cases where all the required components are found
        combinations = self._filtered_db.data[
            ["model", "scenario", "region"]
        ].drop_duplicates()
        for ind in range(len(combinations)):
            model, scenario, region = combinations.iloc[ind]
            found_vars = self._filtered_db.filter(
                model=model, scenario=scenario, region=region
            ).variable
            if any(comp not in found_vars for comp in components):
                self._filtered_db.filter(
                    model=model, scenario=scenario, keep=False, inplace=True
                )
        if len(self._set_of_units_without_equiv(self._filtered_db)) > 1:
            db_to_generate = convert_units_to_MtCO2_equiv(
                self._filtered_db, metric_name=metric_name
            )
        else:
            db_to_generate = self._filtered_db
        consistent_composite = self._construct_consistent_values(
            aggregate, components, db_to_generate
        )
        self._filtered_db.append(consistent_composite, inplace=True)
        cruncher = TimeDepRatio(self._filtered_db)
        if self._set_of_units_without_equiv(
            to_infill_df
        ) != self._set_of_units_without_equiv(consistent_composite):
            raise ValueError(
                "The units of the aggregate variable are inconsistent between the "
                "input and constructed data. We input {} and constructed {}.".format(
                    self._set_of_units_without_equiv(to_infill_df),
                    self._set_of_units_without_equiv(consistent_composite),
                )
            )
        for leader in components:
            to_add = cruncher.derive_relationship(
                leader, [aggregate], only_consistent_cases=False
            )(to_infill_df)
            try:
                df_to_append.append(to_add, inplace=True)
            except NameError:
                df_to_append = to_add
        return df_to_append



"""
Module for the database cruncher which extends using a linear trend
"""
import datetime
import logging
import warnings

from pyam import IamDataFrame

logger = logging.getLogger(__name__)


class LinearExtender:
    """
    Time projector which extends the timeseries of a variable using a linear trend. You
    can either specify a gradient for the line (possibly zero) or a point in the future.
    """

    def __init__(self, db=None):
        """
        Initialise the time projector. The database is only used to determine the
        required timeperiods to return and can be left out if this is specified later.

        Parameters
        ----------
        db : IamDataFrame or None
            The database to use to determine times to return
        """
        if db:
            self._db = db.copy()
        else:
            self._db = db

    def derive_relationship(self, variable, gradient=None, year_value=None, times=None):
        """
        Derives the function to return a linear trend following from the last datapoint

        Parameters
        ----------
        variable : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|CO2"``).

        gradient : float or None
            The gradient of the variable after its last available datapoint, in the
            emissions units per year. If not provided, year_value must be provided
            instead.

        year_value : None or tuple(int or datetime, float)
            The value of the variable at a given future time, e.g. (2050, 0) to extend
            the data to net zero in 2050. If not provided, gradient must be provided
            instead.

        times : None or list[int or datetime]
            The times to return entries at. Only required if no database was used during
            initalisation.


        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable`` timeseries and returns these timeseries extended until the
            latest time in the infiller database.

        Raises
        ------

        ValueError
            There is no data for ``variable`` in the database.

        """
        if times is None:
            if not self._db:
                raise ValueError(
                    "This function must either be given a list of times or a "
                    "database of completed scenarios"
                )
            times = self._db[self._db.time_col].unique()

        if (gradient is not None) and (year_value is not None):
            raise ValueError("Provide only one of a year_value OR gradient")

        if (gradient is None) and (year_value is None):
            raise ValueError("Provide either a year_value OR gradient")

        if year_value:
            if (not isinstance(year_value, tuple)) or (len(year_value) != 2):
                raise ValueError(
                    "year_value should be a tuple of the year and the value that year."
                )

        def filler(in_iamdf):
            """
            Filler function derived from :obj:`LinearExtender`.

            Parameters
            ----------
            in_iamdf : :obj:`pyam.IamDataFrame`
                Input data to fill data in

            Returns
            -------
            :obj:`pyam.IamDataFrame`
                Filled in data (without original source data)

            Raises
            ------
            ValueError
                "The infiller database does not extend in time past the target "
                "database, so no infilling can occur."
            """
            target_df = in_iamdf.filter(variable=variable)

            if target_df.empty:
                error_msg = "No data for `variable` ({}) in target database".format(
                    variable
                )
                raise ValueError(error_msg)
            infiller_time_col = target_df.time_col
            last_time = max(target_df.data[infiller_time_col])
            try:
                later_times = [time for time in times if time > last_time]
            except TypeError as exc:
                raise TypeError(
                    "The times requested must be in the same format as the time column "
                    "in the input database"
                ) from exc

            if not later_times:
                raise ValueError(
                    "No times requested are later than the times already in "
                    "the database"
                )
            key_timepoint_filter = {infiller_time_col: last_time}

            def get_values_in_key_timepoint(idf):
                # filter warning about empty data frame as we handle it ourselves
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    filtered = idf.filter(**key_timepoint_filter)
                idf = filtered.timeseries()
                if not idf.shape[1] == 1:
                    raise AssertionError(
                        "How did filtering for a single timepoint result in more than "
                        "one column?"
                    )
                return idf.iloc[:, 0]

            target_at_key_time = get_values_in_key_timepoint(target_df)

            output_ts = target_df.timeseries()
            # one_year required so type conversion works properly
            one_year = (
                1 if (infiller_time_col == "year") else datetime.timedelta(days=365)
            )
            if year_value:
                gradient_to_use = (year_value[1] - target_at_key_time) / (
                    (year_value[0] - last_time) / one_year
                )
            else:
                gradient_to_use = gradient

            for time in later_times:
                output_ts[time] = target_at_key_time + gradient_to_use * (
                    (time - last_time) / one_year
                )

            for col in output_ts.columns:
                if col not in later_times:
                    del output_ts[col]

            return IamDataFrame(output_ts)

        return filler



"""
Silicone's custom statistical operations.
"""
import os

import numpy as np
import pandas as pd
import scipy.interpolate
import scipy.stats


def rolling_window_find_quantiles(
    xs, ys, quantiles, nwindows=11, decay_length_factor=1
):
    """
    Perform quantile analysis in the y-direction for x-weighted data.

    Divides the x-axis into nwindows of equal length and weights data by how close they
    are to the center of these windows. Then returns the quantiles of this weighted
    data. Quantiles are defined so that the values returned are always equal to a y-
    value in the data - there is no interpolation. Extremal points are given their full
    weighting, meaning this will not agree with the np.quantiles under uniform weighting
    (which effectively gives 0 weight to min and max values).

    The weighting of a point at :math:`x` for a window centered at :math:`x_0` is:

    .. math::

        w = \\frac{1}{1 + \\left (\\frac{x - x_0}{l_{window}} \\times f_{dl} \\right)^2}

    for :math:`l_{window}` the window width (range of values divided by nwindows -1) and
    :math:`f_{dl}` the decay_length_factor.

    Parameters
    ----------
    xs : np.ndarray, :obj:`pd.Series`
        The x co-ordinates to use in the regression.

    ys : np.ndarray, :obj:`pd.Series`
        The y co-ordinates to use in the regression.

    quantiles : list-like
        The quantiles to calculate in each window

    nwindows : int
        How many points to evaluate between x_max and x_min. Must be > 1.

    decay_length_factor : float
        gives the distance over which the weighting of the values falls to 1/4,
        relative to half the distance between window centres. Defaults to 1.

    Returns
    -------
    :obj:`pd.DataFrame`
        Quantile values at the window centres.

    Raises
    ------
    AssertionError
        ``xs`` and ``ys`` don't have the same shape
    """
    if xs.shape != ys.shape:
        raise AssertionError("`xs` and `ys` must be the same shape")

    if isinstance(quantiles, (float, np.float64)):
        quantiles = [quantiles]

    # min(xs) == max(xs) cannot be accessed via QRW cruncher, as a short-circuit appears
    # earlier in the code.
    if np.equal(max(xs), min(xs)):
        # We must prevent singularity behaviour if all the points have the same x.
        window_centers = np.array([xs[0]])
        decay_length = 1
        if np.equal(max(ys), min(ys)):
            return pd.DataFrame(index=window_centers, columns=quantiles, data=ys[0])

    else:
        # We want to include the max x point, but not any point above it.
        # The 0.99 factor prevents rounding error inclusion.
        step = (max(xs) - min(xs)) / (nwindows - 1)
        decay_length = step / 2 * decay_length_factor
        window_centers = np.arange(min(xs), max(xs) + step * 0.99, step)

    ys, xs = map(np.array, zip(*sorted(zip(ys, xs))))

    results = pd.DataFrame(index=window_centers, columns=quantiles)
    results.columns.name = "window_centers"

    for window_center in window_centers:
        weights = 1.0 / (1.0 + ((xs - window_center) / decay_length) ** 2)
        weights /= sum(weights)

        # We want to calculate the weights at the midpoint of step
        # corresponding to the y-value.
        cumsum_weights = np.cumsum(weights) - 0.5 * weights
        results.loc[window_center, quantiles] = scipy.interpolate.interp1d(
            cumsum_weights,
            ys,
            bounds_error=False,
            fill_value=(ys[0], ys[-1]),
            assume_sorted=True,
        )(quantiles)

    return results


def calc_quantiles_of_data(
    distribution, points_to_quant, smoothing=None, weighting=None, to_quantile=True
):
    """
    Calculates the quantiles of points_to_quant in the distribution of values described
    by distribution. Optionally treats points_to_quant as quantiles and returns the
    values that would lead to them instead.

    Parameters
    ----------
    distribution : pd.Series
        The distribution of values.
    points_to_quant : pd.Series
        The points which we want find: if ``to_quantile`` is ``True`` (default) these are the
        values which we will compare to the distribution, if ``False``, these are the
        quantiles which we want to find.
    smoothing : float or string
        By default, no smoothing is done on the distribution. If a value is
        provided, it is fed into :func:`scipy.stats.gaussian_kde` - see full
        documentation there. In short, if a float is input, we fit a Gaussian kernel
        density estimator with that width to the points. If a string is used, it
        must be either "scott" or "silverman", after those two methods of
        determining the best kernel bandwidth.
    weighting : None or Series
        If a series, must have the same indices as distribution, giving the
        relative weights of each point.
    to_quantile : Bool
        If ``True``, we return the quantiles of the data in points_to_quant. If ``False``, we
        instead treat points_to_quant as the quantiles themselves (they must all be 0-1)
        and return the values in distribution that occur at these quantiles.

    Returns
    -------
    :obj:`np.ndarray`
        An array with one row and a column for each entry in points_to_quant, containing the
        quantiles of these points in order. Or, if to_quantile is ``False``, containing the
        values corresponding to the quantiles points_to_quant.
    """
    if not to_quantile:
        assert all(
            [i <= 1 and i >= 0 for i in points_to_quant]
        ), "Quantiles are not defined outside the range 0-1"
    if weighting is None:
        weighting = pd.Series(np.ones_like(distribution), index=distribution.index)
    else:

        assert isinstance(
            weighting, pd.Series
        ), "The weighting variable should be a Series"
        assert len(weighting) == len(distribution), (
            "There must be the same number of weights as entries in the database. "
            "We have len {} in weights and len {} in distribution.".format(
                len(weighting), len(distribution)
            )
        )
        assert np.array_equal(
            weighting.index.sort_values(), distribution.index.sort_values()
        ), (
            "The entries in the weighting Series are not clearly aligned with the "
            "distribution Series. We have \n {} \n in the weighting and \n {} \n in "
            "the distribution ".format(weighting.index, distribution.index)
        )

    if any(np.isnan(distribution)):
        distribution = distribution[~np.isnan(distribution)]
        weighting = weighting[distribution.index]
    len_lead_not_nan = len(distribution)
    if len_lead_not_nan == 1:
        # If there is only a single value then quantile is not defined
        if to_quantile:
            return np.nan * np.zeros_like(points_to_quant)
        else:
            # all quantiles give the same point
            return distribution[0]
    if len_lead_not_nan == 0:
        raise ValueError("No valid data entered to establish the quantiles.")
    if smoothing:
        smooth_points = scipy.stats.gaussian_kde(
            distribution, smoothing, weights=weighting
        )
        minx = min(distribution)
        maxx = max(distribution)
        tail = (maxx - minx) / 2
        # tile the probability space with 10000 points between the tail ends
        xpts = np.linspace(minx - tail, maxx + tail, 10000)
        smooth_dist = np.cumsum(smooth_points(xpts)) * (xpts[1] - xpts[0])
        # But we only want to return values between the extrema
        xkeep = (xpts >= minx) & (xpts <= maxx)
        xpts = xpts[xkeep]
        smooth_dist = smooth_dist[xkeep]

        if to_quantile:
            return scipy.interpolate.interp1d(
                xpts, smooth_dist, bounds_error=False, fill_value=(0, 1)
            )(points_to_quant)
        else:
            return scipy.interpolate.interp1d(
                smooth_dist, xpts, bounds_error=False, fill_value=(minx, maxx)
            )(points_to_quant)

    dist_sort_ind = np.argsort(distribution)
    distribution = distribution[dist_sort_ind.index[dist_sort_ind]]
    weighting = weighting[dist_sort_ind.index[dist_sort_ind]]
    quant_of_lead_vals = (np.cumsum(weighting) - weighting) / (
        sum(weighting) - weighting
    )
    if any(quant_of_lead_vals > 1) or any(quant_of_lead_vals < 0):
        raise ValueError("Impossible quantiles!")
    if to_quantile:
        return scipy.interpolate.interp1d(
            distribution, quant_of_lead_vals, bounds_error=False, fill_value=(0, 1)
        )(points_to_quant)
    else:
        return scipy.interpolate.interp1d(quant_of_lead_vals, distribution)(
            points_to_quant
        )


def calc_all_emissions_correlations(emms_df, years, output_dir):
    """
    Save csv files of the correlation coefficients and the rank correlation
    coefficients between emissions at specified times.

    This function includes all undivided emissions (i.e. results recorded as
    `Emissions|X`) and CO2 emissions split once (i.e. `Emissions|CO2|X`). It does not
    include Kyoto gases. It will also save the average absolute value of the
    coefficients.

    Parameters
    ----------
    emms_df : :obj:`pyam.IamDataFrame`
        The database to search for correlations between named values

    output_dir : str
        The folder location to save the files.

    years : list[int]
        The years upon which to calculate correlations.

    Files created
    -------------
    "variable_counts.csv" : the number of scenario/model pairs where the emissions
    data occurs.

    "gases_correlation_{year}.csv" : The Pearson's correlation between gases emissions
     in a given year.

    "gases_rank_correlation_{year}.csv" : The Spearman's rank correlation between
    gases in a given year

    "time_av_absolute_correlation_{}_to_{}.csv" : The magnitude of the Pearson's
    correlation between emissions, averaged over the years requested.

    "time_av_absolute_rank_correlation_{}_to_{}.csv" : The magnitude of the Spearman's
     rank correlation between emissions, averaged over the years requested.

    "time_variance_rank_correlation_{}_to_{}.csv" : The variance over time in the rank
     correlation values above.
    """
    assert len(emms_df.region) == 1, "Calculation is for only one region"
    # Obtain the list of gases to examine
    df_gases = (
        emms_df.filter(level=1)
        .filter(variable="Emissions|*")
        .filter(variable="Emissions|Kyoto*", keep=False)
        .append(emms_df.filter(level=2).filter(variable="Emissions|CO2*"))[
            ["variable", "unit"]
        ]
        .drop_duplicates()
        .set_index("variable")
    )
    all_correlations_df = pd.DataFrame(
        index=df_gases.index, columns=df_gases.index, data=0
    )
    all_rank_corr_df = pd.DataFrame(
        index=df_gases.index, columns=df_gases.index, data=0
    )
    all_rank_corr_var_df = pd.DataFrame(
        index=df_gases.index, columns=df_gases.index, data=0
    )

    # Calculate the total amount of data
    var_count_file = "variable_counts.csv"
    var_count = pd.Series(index=df_gases.index, dtype=int)
    for var in df_gases.index:
        var_db = emms_df.filter(variable=var)
        var_count[var] = len(var_db.timeseries())
    var_save_loc = os.path.join(output_dir, var_count_file)
    var_count.to_csv(var_save_loc)
    print("Counted the number of each variable and saved to ".format(var_save_loc))

    for year_of_interest in years:
        # Initialise the tables to hold all parameters between runs
        correlations_df = pd.DataFrame(index=df_gases.index, columns=df_gases.index)
        rank_corr_df = pd.DataFrame(index=df_gases.index, columns=df_gases.index)
        # Check that the list has only one entry for each gas
        assert not any(df_gases.index.duplicated()), "Index contains duplicated entries"
        formatted_df = emms_df.filter(
            variable=df_gases.index, year=year_of_interest
        ).pivot_table(
            ["year", "model", "scenario", "region"], ["variable"], aggfunc="mean"
        )
        formatted_df.replace(r"^\s*$", np.nan, regex=True, inplace=True)
        for x_gas_ind in range(df_gases.count()[0]):
            x_gas = df_gases.index[x_gas_ind]
            for y_gas_ind in range(x_gas_ind + 1, df_gases.count()[0]):
                y_gas = df_gases.index[y_gas_ind]
                # Calculate the correlations. This requires removing NAs
                correlations_df.at[y_gas, x_gas] = formatted_df.corr("pearson").loc[
                    x_gas, y_gas
                ]
                rank_corr_df.at[y_gas, x_gas] = formatted_df.corr("spearman").loc[
                    x_gas, y_gas
                ]
                all_correlations_df.loc[y_gas, x_gas] = all_correlations_df.at[
                    y_gas, x_gas
                ] + abs(correlations_df.loc[y_gas, x_gas]) / len(years)
                all_rank_corr_df.loc[y_gas, x_gas] = all_rank_corr_df.at[
                    y_gas, x_gas
                ] + abs(rank_corr_df.at[y_gas, x_gas]) / len(years)
                all_rank_corr_var_df.loc[y_gas, x_gas] = (
                    all_rank_corr_var_df.at[y_gas, x_gas]
                    + rank_corr_df.at[y_gas, x_gas] ** 2
                )
                # the other parts follow by symmetry
                correlations_df.at[x_gas, y_gas] = correlations_df.at[y_gas, x_gas]
                rank_corr_df.at[x_gas, y_gas] = rank_corr_df.at[y_gas, x_gas]
                all_correlations_df.loc[x_gas, y_gas] = all_correlations_df.at[
                    y_gas, x_gas
                ]
                all_rank_corr_var_df.loc[x_gas, y_gas] = all_rank_corr_var_df.loc[
                    y_gas, x_gas
                ]
                all_rank_corr_df.loc[x_gas, y_gas] = all_rank_corr_df.at[y_gas, x_gas]
            print("Finished x_gas {} in year {}.".format(x_gas, year_of_interest))
        if output_dir is not None:
            correlations_df.to_csv(
                os.path.join(
                    output_dir, "gases_correlation_{}.csv".format(year_of_interest)
                )
            )
            rank_corr_df.to_csv(
                os.path.join(
                    output_dir, "gases_rank_correlation_{}.csv".format(year_of_interest)
                )
            )
    # Complete variance calc by removing mean and dividing through
    all_rank_corr_var_df = (
        all_rank_corr_var_df - len(years) * all_rank_corr_df ** 2
    ) / (len(years) - 1)
    if output_dir is not None:
        all_rank_corr_var_df.to_csv(
            os.path.join(
                output_dir,
                "time_variance_rank_correlation_{}_to_{}.csv".format(
                    min(years), max(years)
                ),
            )
        )
    for gas in df_gases.index:
        all_correlations_df.loc[gas, gas] = np.nan
        all_rank_corr_df.loc[gas, gas] = np.nan
    if output_dir is not None:
        all_correlations_df.to_csv(
            os.path.join(
                output_dir,
                "time_av_absolute_correlation_{}_to_{}.csv".format(
                    min(years), max(years)
                ),
            )
        )
        all_rank_corr_df.to_csv(
            os.path.join(
                output_dir,
                "time_av_absolute_rank_correlation_{}_to_{}.csv".format(
                    min(years), max(years)
                ),
            )
        )



"""
Module for the database cruncher which uses the 'latest time quantile' technique.
"""
import logging
import warnings

import numpy as np
from pyam import IamDataFrame

from silicone.stats import calc_quantiles_of_data
from silicone.utils import _make_weighting_series

logger = logging.getLogger(__name__)


class ExtendLatestTimeQuantile:
    """
    Time projector which extends the timeseries of a variable by assuming that it
    remains that a fixed quantile in the infiller database, the quantile it is in at the
    last available time. This is the natural counterpart to the equal quantile walk
    extending a single variable over time rather than over different emissions.

    It assumes that the target timeseries is shorter than the infiller timeseries.
    """

    def __init__(self, db):
        """
        Initialise the time projector with a database containing data from the full
        range of times you wish to see in the output.

        Parameters
        ----------
        db : IamDataFrame
            The database to use
        """
        self._db = db.copy()

    def derive_relationship(self, variable, smoothing=None, weighting=None):
        """
        Derives the quantiles of the variable in the infiller database. Note that this
        takes only one variable as an argument, whereas most crunchers take two.

        Parameters
        ----------
        variable : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|CO2"``).

        smoothing : float or string
            By default, no smoothing is done on the distribution. If a value is
            provided, it is fed into :func:`scipy.stats.gaussian_kde` - see full
            documentation there. In short, if a float is input, we fit a Gaussian kernel
            density estimator with that width to the points. If a string is used, it
            must be either "scott" or "silverman", after those two methods of
            determining the best kernel bandwidth.

        weighting : None or dict{(str, str): float}
            The dictionary, mapping the (model and scenario) tuple onto the weight (
            relative to a weight of 1 for the default). This does not have to include
            all scenarios in df, but cannot include scenarios not in df.

        Returns
        -------
        :obj:`func`
            Function which takes a :obj:`pyam.IamDataFrame` containing
            ``variable`` timeseries and returns these timeseries extended until the
            latest time in the infiller database.

        Raises
        ------

        ValueError
            There is no data for ``variable`` in the database.

        """
        iamdf = self._get_iamdf_variable(variable)

        infiller_time_col = iamdf.time_col
        data_follower_unit = iamdf.data["unit"].unique()

        assert (
            len(data_follower_unit) == 1
        ), "The infiller database has {} units in it. It should have one. ".format(
            len(data_follower_unit)
        )
        if not isinstance(weighting, type(None)):
            if type(weighting) == dict:
                weighting = _make_weighting_series(iamdf.timeseries(), weighting)
            else:
                raise ValueError("We can only use dictionary values for weights")

        def filler(in_iamdf):
            """
            Filler function derived from :obj:`LatestTimeRatio`.

            Parameters
            ----------
            in_iamdf : :obj:`pyam.IamDataFrame`
                Input data to fill data in

            Returns
            -------
            :obj:`pyam.IamDataFrame`
                Filled in data (without original source data)

            Raises
            ------
            ValueError
                "The infiller database does not extend in time past the target "
                "database, so no infilling can occur."
            """
            target_df = in_iamdf.filter(variable=variable)
            if target_df.empty:
                error_msg = "No data for `variable` ({}) in target database".format(
                    variable
                )
                raise ValueError(error_msg)
            if infiller_time_col != in_iamdf.time_col:
                raise ValueError(
                    "`in_iamdf` time column must be the same as the time column used "
                    "to generate this filler function (`{}`)".format(infiller_time_col)
                )

            key_timepoint = max(target_df.data[infiller_time_col])
            later_times = [
                t for t in iamdf.data[infiller_time_col].unique() if t > key_timepoint
            ]
            if not later_times:
                raise ValueError(
                    "The infiller database does not extend in time past the target "
                    "database, so no infilling can occur."
                )
            key_timepoint_filter = {infiller_time_col: key_timepoint}

            def get_values_in_key_timepoint(idf):
                # filter warning about empty data frame as we handle it ourselves
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    filtered = idf.filter(**key_timepoint_filter)
                idf = filtered.timeseries()
                if not idf.shape[1] == 1:
                    raise AssertionError(
                        "How did filtering for a single timepoint result in more than "
                        "one column?"
                    )
                return idf.iloc[:, 0]

            infiller_at_key_time = get_values_in_key_timepoint(iamdf)

            target_at_key_time = get_values_in_key_timepoint(target_df)

            quantiles = calc_quantiles_of_data(
                infiller_at_key_time, target_at_key_time, smoothing, weighting
            )
            if any(np.isnan(quantiles)):
                logger.warning("Only a single value provided for calculating quantiles")
                quantiles = [0.5 if np.isnan(q) else q for q in quantiles]
            output_ts = target_df.timeseries()
            iamdf_ts = iamdf.timeseries()
            for time in later_times:
                output_ts[time] = calc_quantiles_of_data(
                    iamdf_ts[time], quantiles, smoothing, weighting, to_quantile=False
                )
            for col in output_ts.columns:
                if col not in later_times:
                    del output_ts[col]
            return IamDataFrame(output_ts)

        return filler

    def _get_iamdf_variable(self, variable):
        if variable not in self._db.variable:
            error_msg = "No data for `variable` ({}) in database".format(variable)
            raise ValueError(error_msg)

        return self._db.filter(variable=variable)



"""
Uses the same cruncher (by default the 'quantile rolling windows') to split an aggregate
emission into all but one of its constituents and infills the remainder as another
specified emissions type (which may be negative).
"""

import logging

from silicone.database_crunchers import QuantileRollingWindows
from silicone.multiple_infillers import infill_composite_values
from silicone.utils import _remove_equivs, convert_units_to_MtCO2_equiv

logger = logging.getLogger(__name__)


class SplitCollectionWithRemainderEmissions:
    """
    Splits the known aggregate emissions into several components with a specified
    cruncher (by default the 'quantile rolling windows' cruncher), then allocates any
    remainder to ``remainder``.
    """

    def __init__(self, db):
        """
        Initialises the database to use for infilling.

        Parameters
        ----------
        db : IamDataFrame
            The database for infilling.
        """
        self._db = db.copy()

    def _check_and_return_desired_unit(
        self, relevant_df, aggregate, components, remainder
    ):
        """
        Converts the units of the component emissions to be the same as the
        aggregate emissions. Returns the converted database and the unit.

        Parameters
        ----------
        relevant_df : :obj:`pyam.IamDataFrame`
            Data with units that need correcting.

        aggregate : str
            The name of the aggregate variable.

        components : [str]
            List of the names of the variables to be summed.

        remainder : str
            The component which will be constructed as a remainder.

        Return
        ------
        :obj:`pyam.IamDataFrame`
            Data with consistent units.

        str
            The unit of the aggregate data.

        Raises
        ------
        ValueError
            The variables in this dataframe have units that cannot easily be converted
            to make them consistent.
        """
        all_var = [aggregate, remainder] + components
        all_units = relevant_df.data[["variable", "unit"]].drop_duplicates()
        if not all(var in all_units["variable"].values for var in all_var):
            logger.warning(
                "Some variables missing from database when performing "
                "unit conversion: {}".format(
                    [var for var in all_var if var not in all_units["variable"]]
                )
            )
        assert (
            aggregate in all_units["variable"].values
        ), "No aggregate data in database."
        if remainder:
            assert (
                remainder in all_units["variable"].values
            ), "No remainder data in database."
        desired_unit = all_units["unit"][all_units["variable"] == aggregate]
        assert len(desired_unit) == 1, "Multiple units for the aggregate variable"
        desired_unit = desired_unit.iloc[0]
        desired_unit_eqiv = _remove_equivs(desired_unit)
        unit_equivs = all_units["unit"].map(_remove_equivs).drop_duplicates()
        if len(unit_equivs) != 1 and desired_unit_eqiv != "Mt CO2/yr":
            raise ValueError(
                "The variables in this dataframe have units that cannot "
                "easily be converted to make them consistent."
            )
        return relevant_df, desired_unit

    def infill_components(
        self,
        aggregate,
        components,
        remainder,
        to_infill_df,
        cruncher_class=QuantileRollingWindows,
        metric_name="AR5GWP100",
        **kwargs,
    ):
        """
        Derive the relationship between the composite variables and their sum, then use
        this to deconstruct the sum.

        Parameters
        ----------
        aggregate : str
            The variable for which we want to calculate timeseries (e.g.
            ``"Emissions|CO2"``).

        components : list[str]
            The variables to be infilled directly by the cruncher. (e.g.
            ``["Emissions|CO2|AFOLU", "Emissions|CO2|Energy"]``).

        remainder : str
            The variable which will absorb any difference between the aggregate and
            component emissions. This may be positive or negative. E.g.
            ``"Emissions|CO2|Industry"``

        to_infill_df : :obj:`pyam.IamDataFrame`
            The dataframe that already contains the ``aggregate`` variable, but needs
            the ``components`` to be infilled.

        cruncher_class : :class:
            The cruncher used to perform the infilling. By default this will be
            QuantileRollingWindows.

        metric_name : str
            The name of the conversion metric to use. This will usually be
            AR<4/5/6>GWP100.

        **kwargs :
            Handed to the cruncher.

        Returns
        -------
        :obj:`pyam.IamDataFrame`
            The infilled data resulting from the calculation.
        """
        assert (
            aggregate in to_infill_df.variable
        ), "The database to infill does not have the aggregate variable"
        assert all(
            y not in [remainder] + components for y in to_infill_df.variable
        ), "The database to infill already has some component variables"
        assert len(to_infill_df.data.columns) == len(self._db.data.columns) and all(
            to_infill_df.data.columns == self._db.data.columns
        ), (
            "The database and to_infill_db fed into this have inconsistent columns, "
            "which will prevent adding the data together properly."
        )
        to_infill_df = to_infill_df.filter(variable=aggregate)
        assert (
            len(to_infill_df["unit"].unique()) == 1
        ), "Multiple units in the aggregate data"
        to_infill_ag_units = to_infill_df.unit[0]
        all_var = [aggregate, remainder] + components
        relevant_df = self._db.filter(variable=all_var)
        db_to_generate, aggregate_unit = self._check_and_return_desired_unit(
            relevant_df, aggregate, components, remainder
        )
        assert _remove_equivs(aggregate_unit) == _remove_equivs(to_infill_ag_units), (
            "The units of the aggregate variable are different between infiller and "
            "infillee dataframes"
        )
        cruncher = cruncher_class(db_to_generate)
        unavailable_comp = [
            var for var in components if var not in relevant_df.variable
        ]
        if unavailable_comp:
            logger.warning("No data found for {}".format(unavailable_comp))
        remainder_dict = {aggregate: 1}
        components = [comp for comp in components if comp not in unavailable_comp]
        for leader in components:
            to_add = cruncher.derive_relationship(leader, [aggregate], **kwargs)(
                to_infill_df
            )
            try:
                df_to_append.append(to_add, inplace=True)
            except NameError:
                df_to_append = to_add
            remainder_dict[leader] = -1
        # We need a single database with both aggregate and components, which we may
        # want to convert units to CO2/yr before we calculate the remainder.
        calculate_remainder_df = df_to_append.append(to_infill_df)
        if _remove_equivs(aggregate_unit) == "Mt CO2/yr":
            calculate_remainder_df = convert_units_to_MtCO2_equiv(
                calculate_remainder_df, metric_name
            )

        df_to_append.append(
            infill_composite_values(
                calculate_remainder_df, {remainder: remainder_dict}
            ),
            inplace=True,
        )
        return df_to_append
